\\ Copyright 2014, 2015, 2017 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


print("GP version() = ",version)

have_strictargs = version[1]>2 || (version[1]==2 && version[2]>=7);
print("have_strictargs = ",have_strictargs);
if(have_strictargs, default(strictargs,1));

read("memoize.gp");

test_count = 0;
test_equal(got,want,name="")=
{ test_count++;
  if(got==want,
     \\ print("ok ",name)
     ,
     print("wrong - ",name"\ngot  = ",got,"\nwant = ",want);
     quit(1));
}

\\ have_variadic =
\\ {
\\   if(iferr=='iferr,
\\      eval("trap(,0,eval(\"(v[..])->v\")||1)"),
\\      eval("iferr(eval(\"(v[..])->v\")||1, e,0)"));
\\ }
\\
have_variadic = version[1]>2 || (version[1]==2 && version[2]>=8);
print("have_variadic = ",have_variadic);


print ("memoize_INTERNAL_args = ",memoize_INTERNAL_args);
print ("memoize_INTERNAL_maps = ",memoize_INTERNAL_maps);

\\-------------------------------
\\ memoize_INTERNAL_any_comma()

test_equal(memoize_INTERNAL_any_comma(""), 0,  "empty string")
test_equal(memoize_INTERNAL_any_comma("a"), 0, "not empty string")
test_equal(memoize_INTERNAL_any_comma("a,b"), 1, "a,b")
test_equal(memoize_INTERNAL_any_comma(",b"), 1)
test_equal(memoize_INTERNAL_any_comma("b,"), 1)

\\-------------------------------

zero_args()="ret zero";
test_equal(memoize_INTERNAL_func_args(zero_args), ["","",0])
test_equal(memoize_INTERNAL_func_is_zero_args(zero_args), 1)

one_args(n)="ret one";
test_equal(memoize_INTERNAL_func_args(one_args), ["n","n",0])
test_equal(memoize_INTERNAL_func_is_zero_args(one_args), 0)

two_args(abc0,def1)="ret two";
test_equal(memoize_INTERNAL_func_args(two_args), ["abc0,def1","abc0,def1",0])
test_equal(memoize_INTERNAL_func_is_zero_args(two_args), 0)

default_expr(x=(1+(2+3)+4),y) = x+y;
test_equal(memoize_INTERNAL_func_args(default_expr), ["x,y","x=(1+(2+3)+4),y",0])
test_equal(memoize_INTERNAL_func_is_zero_args(default_expr), 0)

help_example(x,y=2*(x+1)) = 3*x+4*y;
test_equal(memoize_INTERNAL_func_args(help_example), ["x,y", "x,y=2*(x+1)",0])

default_str(x="(\"\\\\") = x+y;
test_equal(memoize_INTERNAL_func_args(default_str), ["x", "x=\"(\\\"\\\\\\\\\"",0])

{
  if(have_variadic,
     eval("var_args(v[..]) = #v;");
     test_equal(var_args(9,10,11), 3);
     test_equal(memoize_INTERNAL_func_args(var_args), ["v","v[..]",1]);
     test_equal(memoize_INTERNAL_func_is_zero_args(var_args), 0));
}
{
  if(have_variadic,
     eval("var_args(a,b=2,v[..]) = #v;");
     test_equal(var_args(9,10,11), 1, "var_args() 2 fixed");
     test_equal(memoize_INTERNAL_func_args(var_args),["a,b,v","a,b=2,v[..]",1]);
     test_equal(memoize_INTERNAL_func_is_zero_args(var_args), 0));
}


\\-------------------------------
foo_called=0;
foo_value=123;
foo(n) = foo_called++; foo_value;
foo=memoize(foo);
\\ print("memoize(foo) is ",foo);

test_equal(foo(1), 123)
test_equal(foo_called, 1)
test_equal(foo(1), 123)
test_equal(foo_called, 1)

foo_value=456;
test_equal(foo(1), 123)
test_equal(foo("abc"), 456)
test_equal(foo_called, 2)
foo_value="blah";

\\-------------------------------
bar_called=0;
bar_value=789;
bar(n : small) = bar_called++; bar_value;
bar=memoize(bar);

test_equal(bar(-1.5), 789)
test_equal(bar_called, 1)

bar_value=32;
test_equal(bar(-1.5), 789)
test_equal(bar(1.5), 32)
test_equal(foo(1), 123)
test_equal(foo("abc"), 456)
test_equal(foo_called, 2)
test_equal(bar_called, 2)

\\-------------------------------
triplet_called=0;
triplet_value=10;
triplet(a,b,c) = triplet_called++; triplet_value;
triplet=memoize(triplet);

test_equal(triplet(1,2,3), 10)
test_equal(triplet_called, 1)

triplet_value=20;
test_equal(triplet(1,2,3), 10)
test_equal(triplet_called, 1)
test_equal(triplet(2,1,3), 20)
test_equal(triplet_called, 2)


\\-------------------------------
defer_called=0;
defer_value=[1,2,3];
defer() = defer_called++; defer_value;
defer=memoize(defer);

test_equal(defer(), [1,2,3])
defer_value=99;
test_equal(defer(), [1,2,3])
test_equal(defer_called, 1)


\\-------------------------------
fib=memoize(n -> if(n<2,n, fib(n-1)+fib(n-2)));
test_equal(fib(10), fibonacci(10))


\\-------------------------------
defaulter=memoize((x,y=x+1)->10*x+y);
test_equal(defaulter(1), 12)
test_equal(defaulter(2), 23)
test_equal(defaulter(2,5), 25)


\\-------------------------------
ff(func) = 2*func;
ff=memoize(ff);
test_equal(ff(1), 2)
test_equal(ff(10), 20)


\\-------------------------------
dummy(x) = x;
foo = memoize(n -> dummy(n); dummy(2*n); 3*n);
test_equal(foo(4), 12)
foo = (n -> dummy(n); dummy(2*n); 3*n);
test_equal(foo(4), 12)


\\-------------------------------
\\ Cannot memoize an install() function

expected_install_error = \
 "cannot memoize an install()ed function (put a gp wrapper around it)";

install(time, "lf", "t");
test_equal(memoize_INTERNAL_func_is_zero_args(t), 0)
test_equal(iferr(memoize_INTERNAL_func_args(t), e, component(e,1)[1], "t_USER"), \
           expected_install_error)      

install(time, "lf");
test_equal(memoize_INTERNAL_func_is_zero_args(time), 0)
test_equal(iferr(memoize_INTERNAL_func_args(time), e, component(e,1)[1], "t_USER"), \
           expected_install_error)      


\\-------------------------------
\\ Can wrap an install function in a gp level function

install(time, "lf", "C_library_time_t");

wtime() = C_library_time_t();     \\ gp front end
wtime = memoize(wtime);

test_equal(wtime(), wtime())

\\ per the comments in memoize.gp
install(time, "lf", "C_library_time__via_install");
C_library_time() = C_library_time__via_install();
C_library_time = memoize(C_library_time);
test_equal(C_library_time(), C_library_time());

C_library_time = memoize(() -> C_library_time__via_install());
test_equal(C_library_time(), C_library_time());


\\-------------------------------
\\ Errors are not memoized.

my_err(n) = if(n==2, error("n=2 is an error")); 4*n;
test_equal(my_err(3), 12);
my(caught=0); iferr(my_err(2),err, caught=1); test_equal(caught,1);
my(caught=0); iferr(my_err(2),err, caught=1); test_equal(caught,1);

my_err_noargs() = error("no args error");
my(caught=0); iferr(my_err_noargs(),err, caught=1); test_equal(caught,1);
my(caught=0); iferr(my_err_noargs(),err, caught=1); test_equal(caught,1);


\\-------------------------------
\\ variadic memoized
foo_called=0;
foo_value=123;
{
  if(have_variadic,
     eval("foo(v[..]) = foo_called++; foo_args=v; foo_value;");
     foo = memoize(foo);
     \\ print(foo);

     test_equal(foo(1), 123, "variadic ret");
     test_equal(foo_called, 1);
     test_equal(foo_args, [1], "variadic args");

     foo_called = 0;
     test_equal(foo(1), 123);      \\ second call
     test_equal(foo_called, 0, "variadic no second call");

     test_equal(foo(6,7,8), 123, "variadic3 ret");
     test_equal(foo_called, 1);
     test_equal(foo_args, [6,7,8]);

     foo_called = 0;
     test_equal(foo(6,7,8), 123);  \\ second call
     test_equal(foo_called, 0);
    );
}

foo=memoize(foo);
\\ print("memoize(foo) is ",foo);


\\-------------------------------
print(test_count," tests ok");
